/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */
/* 
   file     : ed2k_data.c
   purpose  : internal data management of files, users, and searches in these

*/


/*
  ATTENTION: 
  be careful when working with mutexes. all functions beginning
  with ed2k_data_ are NOT to be called from the outside. so these
  are only helper functions for others. 
  mutexes are to be realized in the functions being called from 
  the outside, this is much safer.
*/


#include <glib-2.0/glib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "littlehelpers.h"
#include "maintenance.h"
#include "structs.h"
#include "ed2k_message_gen.h"
#include "settings.h"

/* globals */



/* ************************************************************************** */
/*                      trees and their locks                                 */
/* ************************************************************************** */


GTree   *FileTree; 
GTree   *UserTree;
GTree   *DescTree; 
GTree   *WordTree;

GMutex  *rwlock;


guchar **welcome_message;

extern struct central_struct *rome;



/* ************************************************************************** */
/*                        some global variables                               */
/* ************************************************************************** */


gint                      num_users;
gint                      num_users_lowid;
gint                      num_files_all;
gint                      num_files_different;
gint                      num_servers;
gpointer                 *low_ids;
gint                      size_all_files;



/* ************************************************************************** */
/* some neccessary definitions (to be replaceed by ints later)                */
/* ************************************************************************** */


#define MAX_LOW_ID_COUNT        1000




/* ************************************************************************** */
/* NOW THE REAL STUFF                                                         */
/* ************************************************************************** */



/* ************************************************************************** */
/*                               internals                                    */
/* ************************************************************************** */

struct UserInfo *
search_user_id_internal(guint id)
{
   
   gboolean found;
   guint *found_id;
   struct UserInfo *user;
   
   found = g_tree_lookup_extended(UserTree, &(id), 
				  (gpointer) &found_id, 
				  (gpointer) &user);

   if (!found) 
      return NULL;
   
   g_assert(id == *found_id);
   g_assert(id ==  user->id);

   return user;
}




/* ************************************************************************** */
/*                        some debug functions                                */
/* ************************************************************************** */




gboolean show_users(gpointer user_id, struct UserInfo *user, gpointer p2)
{
   printf("\tuser pointer: %d, files: %4d, user->name: %s\n", 
	  (unsigned int) user, g_tree_nnodes(user->shared_files), user->name);
   return FALSE;
}


gboolean show_words(char *word, GTree *descs, gpointer p2)
{
   printf("\t%20s\t%2d descriptions\n", word, g_tree_nnodes(descs));
   return FALSE;
}


gboolean show_descriptions(char *string, struct DescInfo *desc, gpointer p2)
{
   printf("\t(%2d files) \"%s\"\n", g_tree_nnodes(desc->files), string);
   return FALSE;
}


gboolean show_file_owners(struct UserInfo *owner, struct DescInfo *desc, gpointer np)
{
   printf("  %.15s : %s\n", owner->name, desc->string);
   return FALSE;
}


gboolean show_file(char *hash, struct FileInfo *file, int *np)
{
   printf("file no. %3d\n", *np);
   g_tree_foreach(file->owners, 
		  (GTraverseFunc) show_file_owners, NULL);
   *np = *np + 1;
   return FALSE;
}


void debug_dump()
{
   int *count;
   count = g_new0(int, 1);
   printf("remaining files: %d\n", g_tree_nnodes(FileTree));
   printf("remaining users: %d\n", g_tree_nnodes(UserTree));
   g_tree_foreach(UserTree, (GTraverseFunc)show_users, NULL);
   printf("remaining words (%d):\n", g_tree_nnodes(WordTree));
   g_tree_foreach(WordTree, (GTraverseFunc)show_words, NULL);
   printf("remaining descriptions: %d\n", g_tree_nnodes(DescTree));
   g_tree_foreach(DescTree, (GTraverseFunc)show_descriptions, NULL);

   g_tree_foreach(FileTree, (GTraverseFunc)show_file, count);

   printf("remaining users: %d\n", g_tree_nnodes(UserTree));
   printf("remaining files: %d\n", g_tree_nnodes(FileTree));
   printf("remaining descriptions: %d\n", g_tree_nnodes(DescTree));
   printf("remaining words (%d):\n", g_tree_nnodes(WordTree));
}







/* ************************************************************************** */
/* functions neccessary for the outside for comfort                           */
/* ************************************************************************** */



int ed2k_data_free_search_results(struct SearchResults *results)
{
   struct SearchResults *next;

   while (results) {
      next = results->next;
      g_free(results);
      results = next;
   }

   return 1;
}


int ed2k_data_free_message(struct message_handle_struct *message)
{
   if (message->udp_packet)
      {
	 gnet_udp_packet_delete(message->udp_packet);
	 g_free(message->udp_address);
      }

   g_free(message->buffer);
   g_free(message);

   
   return 1;
}


/* ************************************************************************** */
/* statistics & informations                                                  */
/* ************************************************************************** */


int ed2k_data_count_file_owners(struct FileInfo *file)
{
   return g_tree_nnodes(file->owners);
}


int ed2k_data_count_different_files()
{
   return g_tree_nnodes(FileTree);
}
    

int ed2k_data_count_user_files(struct UserInfo *user)
{
   return g_tree_nnodes(user->shared_files);
}


int ed2k_data_count_users()
{
   return g_tree_nnodes(UserTree);
}


// TODO: reimplement??
int ed2k_data_count_all_files();


int ed2k_data_get_all_files_count()
{
   return num_files_all;
}

int ed2k_data_get_different_files_count()
{
#ifdef debug
   if (g_tree_nnodes(FileTree) != num_files_different);
   //g_assert_not_rached();
#endif
   return num_files_different;
}

int ed2k_data_get_users_count()
{
   return num_users;
}


int ed2k_data_get_lowid_count()
{
   return num_users_lowid;
}


gint
ed2k_data_get_files_byte_count()
{
   return size_all_files;
}



/* ************************************************************************** * 
 * searching and moving stuff ...  :-))                                       * 
 * ************************************************************************** */



/* ************************************************************************** * 
 * low id stuff                                                               *
 * ************************************************************************** */


gint
get_next_low_id()
{
   gpointer *id_parse;
   int   i;
   
   i = 0;

   id_parse = low_ids;
   while (id_parse[i])
      i++;

   return i+1;
}


gint
reserve_low_id(gint id, struct UserInfo *user)
{
   if (!id) 
      return -1;

   id--;

   if (id >= MAX_LOW_ID_COUNT)
      return -1;
   
   if ( low_ids[id] )
      {
	 // already used
	 return 0;
      }
   else 
      {
	 // return reserved id
	 low_ids[id] = user;

	 return id+1;
      }

   g_assert_not_reached();
   return 1;
}




gint
free_low_id(gint id)
{
   if (!id)
      return -1;
   
   id--;

   if (id >= MAX_LOW_ID_COUNT)
      return -1;
   
   if (low_ids[id] == NULL)
      return 0;
   else
      {
	 low_ids[id] = NULL;
#ifdef debug
	 printf("\tlow_id %d freed\n", id+1);
#endif
	 return id+1;
      }
      
   g_assert_not_reached();
   return 1;
}





/* ************************************************************************** * 
 * sort functions for the trees                                               *
 * ************************************************************************** */





gint 
simple_order_func(gpointer  p1, gpointer p2)
{
   if (p1 < p2)
      return -1;
   else if (p1 > p2)
      return 1;
   else 
      return 0;
}



int
user_id_order_func(guint *id1, guint *id2)
{
   if (*id1 < *id2) 
      return -1;
   else if (*id1 > *id2)
      return 1;
   else return 0;
}



int 
filehash_order_func(struct FileInfo *p1, struct FileInfo *p2)
{
   // could this be implemented without this func call ??? 
   // ... i think yes .. 
   
   return memcmp(p1, p2, 16);
}


int client_id_order_func(gpointer user1, gpointer user2)
{
   return memcmp(user1, user2, 4);
}









/* ************************************************************************** * 
 * functions to clean up behind to-be-removed things                          *
 * ************************************************************************** */



int file_remove_finally(struct FileInfo *file)
{
   // now the files should not have any owners left. 
   // so just delete it :-)

   // TODO: logging
   
   if (file->add_info)
      g_free(file->add_info);
   
   // decrease the size of the files hold by the server
   size_all_files -= file->size;

   g_tree_destroy(file->owners);
   
   g_free(file);

   return 1;
}


int description_remove_finally(struct DescInfo *desc)
{
   // this is only called if the description refers to no more files.

   gchar   **doubles,
           **words,
            *lower_string,
            *orig_word; 
   guint     i;
   gboolean  found_word;
   GTree    *words_tree;


   // TODO: (error) logging 

   // TODO: word things (check & remove if neccessary)
   
   lower_string = g_ascii_strdown(desc->string, -1);
   doubles = parse_string(lower_string, -1);
   words = remove_doubles(doubles);

   i = 0;
   while (words[i])
   {

	 // first: lookup word
	 found_word = g_tree_lookup_extended(WordTree, 
					     words[i], 
					     (gpointer*) &orig_word,
					     (gpointer*) &words_tree);

	 if (!found_word)
	    printf("\n\n W O R D   %s   N O T    F O U N D ! ! ! ! ! \n\n", words[i]);
	 g_assert(found_word);
	 
	 // second: remove description from words desc tree
	 g_tree_remove(words_tree, desc);

	 // third: if the word is in no more descriptions - delete
	 if (g_tree_nnodes(words_tree) == 0) 
	    {
	       g_tree_remove(WordTree, words[i]);
	       g_free(orig_word);
	       g_tree_destroy(words_tree);
	    }

	 i++;
   }

   // FUTURE: keep deletion delayed for thread safe reasons ... 

   free_parsed_strings(words);    
   g_tree_destroy(desc->files);   
   g_free(desc->string);
   g_free(desc);


   return 1;
}




struct DescInfo *description_increase_ref_count(char *desc, 
						short desc_length,
						struct FileInfo *file)
{
   // this function does also add the desc.
   // if its not already present

   // TODO: LOCKING & UNLOCKING ??

   char             *d_string;
   char             *description;
   char             *new_word;
   struct DescInfo  *d_struct;
   struct FileInfo  *orig_file;
   gint              file_count;
   gboolean          file_already_present,
                     desc_already_present,
                     word_already_present;
   gchar           **raw_words,
                   **words,
                    *orig_word,
                    *lower_string;
   GTree            *words_tree,
                    *orig_words_tree;
   guint             i;
   guint             word_len;




   description = g_new0(char, desc_length+1);
   if (desc_length == -1)
      desc_length = strlen(desc);
   memcpy(description, desc, desc_length);


   desc_already_present = g_tree_lookup_extended(DescTree, 
						 description, 
						 (gpointer*) &d_string,
						 (gpointer*) &d_struct);
	 
   if (! desc_already_present) 
      {
	 
	 d_struct = g_new0(struct DescInfo, 1);
	 d_struct->string = description;
	 d_struct->strlen = desc_length;
	 d_struct->files = g_tree_new((GCompareFunc)simple_order_func);
	 
	 g_tree_insert(DescTree, description, d_struct);
	 
	 
	 lower_string = g_ascii_strdown(description, -1);
	 raw_words = parse_string(lower_string, -1);
	 words = remove_doubles(raw_words);
	 g_free(lower_string);
	 
	 
	 i = 0;
	 while (words[i])
	    {
	       word_already_present = 
		  g_tree_lookup_extended(WordTree,
					 words[i],
					 (gpointer*) &orig_word,
					 (gpointer*) &orig_words_tree);
	       
	       if (word_already_present) 
		  {
		     // just add the new description to the word
		     g_tree_insert(orig_words_tree, d_struct, NULL);
		  }
	       else
		  {
		     // first:  duplicate the word
		     // second: create binary tree for the word
		     // third:  insert the description into the words' tree
		     // fourth: insert word into global words tree
		     
		     // first
		     word_len = strlen(words[i]);
		     new_word = g_new0(char, word_len+1);
		     memcpy(new_word, words[i], word_len);
		     // second
		     words_tree = g_tree_new((GCompareFunc)simple_order_func);
		     // third
		     g_tree_insert(words_tree, d_struct, NULL);
		     //fourth
		     g_tree_insert(WordTree, new_word, words_tree);
		  }
	       
	       i++;
	    }
	 
	 free_parsed_strings(words);
	 
	 // dont forget to insert the file into the desc files tree!
	 g_tree_insert(d_struct->files, file, (gpointer) 1);
	 
	 d_struct->ref_count = 1;
	 
      }
   else
      {
	 file_already_present = g_tree_lookup_extended(d_struct->files,
						       file, 
						       (gpointer*) &orig_file,
						       (gpointer*) &file_count);
	 
	 if (!file_already_present) 
	    {
	       // add with count of 1
	       g_tree_insert(d_struct->files, file, (gpointer) 1);
	    }
	 else
	    {
	       // add and increase count
	       g_tree_insert(d_struct->files, file, (gpointer)(file_count + 1));
	    }
	 
	 g_free(description);
	 d_struct->ref_count++;
      }
   
   
   

   return d_struct;
}





struct DescInfo *description_decrease_ref_count(struct DescInfo *desc, 
						struct FileInfo *file)
{
   struct FileInfo *orig_file;
   int              file_count;
   int              files_of_desc;
   struct DescInfo *returnme;
   

   returnme = desc;
   
   // BULLSHIT: dont remove the file. decrease the ref_count.
   g_tree_lookup_extended(desc->files, file, 
			  (gpointer*) &orig_file, 
			  (gpointer*) &file_count);

   // ONLY remove if the desc points NO LONGER to THIS SPEICAL file
   if (file_count == 1) 
   {
	 // if the desc has no more files left

	 g_tree_remove(desc->files, file);
   }
   else
   {
	 desc->ref_count--;
	 g_tree_insert(desc->files, file, (gpointer)(file_count - 1));
   }


   // and ONLY remove desc if it NO LONGER points to ANY file

   files_of_desc = g_tree_nnodes(desc->files);
   if (files_of_desc == 0) 
      {

	 files_of_desc = g_tree_nnodes(desc->files);
	 if (files_of_desc == 0) 
	    {
	       // TODO: WRITE LOCK & SECOND CHECK
	       	       
	       // not harmful, cause nothing is freed! data pointers remain valid
	       g_tree_remove(DescTree, desc->string);
	       description_remove_finally(desc);
	       returnme = NULL;
	    }
      }
      

   return returnme;
}





// is called iteratively. locked is before and after the call ... 
int 
iteration_file_owner_remove(struct FileInfo *file, 
		      struct DescInfo *desc, 
		      struct UserInfo *owner)
{
   gint owners;
   struct UserInfo *orig_owner;
   struct DescInfo *orig_desc;
   

   // first take care of the description thing ...
   // lookup original owner->description pair


   g_tree_lookup_extended(file->owners, owner, 
			  (gpointer*) &orig_owner, 
			  (gpointer*) &orig_desc);

   description_decrease_ref_count(orig_desc, file);

   // now finally remove the owner
   g_tree_remove(file->owners, owner);
   file->count--;

   // if no more owners left delete file
   owners = g_tree_nnodes(file->owners);
   
   if (owners != file->count)
   {
	 printf("\n\n  E R R O R   -- data integrity error\n");
	 printf("              -- files count does not match!!\n\n");
   }

   if (owners == 0) 
      {      
	 g_tree_remove(FileTree, file->hash);
	 num_files_different--;
	 file_remove_finally(file);
      }
   
   num_files_all--;
   

   return FALSE;
}





gboolean 
iteration_generate_dl_sources(struct UserInfo *user, 
			      struct DescInfo *desc,
			      struct SearchBaseStruct *sstruct)
{
   struct SearchResults *source;

   // TODO: implement some nice rotation scheme here!

   source           = g_new(struct SearchResults, 1);
   source->user     = user;
   source->next     = sstruct->results;
   sstruct->results = source;


   return FALSE;
}






gboolean 
iteration_add_desc_files(struct FileInfo *file,
			 gpointer file_count, 
			 struct SearchBaseStruct *sstruct)
{
   struct SearchResults *temp;
   
   g_assert(sstruct->current_desc); // hope this will work ... 
   // well, yes.

   temp = sstruct->results;
   sstruct->results = g_new0(struct SearchResults, 1);
   
   sstruct->results->file   = file; 
   sstruct->results->desc   = sstruct->current_desc;
   sstruct->results->next   = temp;


   return FALSE;
}





gboolean iteration_search_desc(struct DescInfo *d_struct, 
			       gpointer gnull, 
			       struct SearchBaseStruct *sstruct)
{
      
   gpointer found_null;
   struct DescInfo *found_desc;
   int cmp;

   cmp = 1;
   while (sstruct->strees[cmp])
      cmp++;
   
   if (ndl(5))
      printf("\t search: in description: %s\n", d_struct->string);

   // this is not ax expression. this is a monster ;-)
   if (cmp == 1)
      {
	 sstruct->current_desc = d_struct;
	 g_tree_foreach(d_struct->files, 
			(GTraverseFunc) iteration_add_desc_files, sstruct);
      }
   else if (    ( ( cmp == 2) && g_tree_lookup_extended(sstruct->strees[1], 
							d_struct, 
							(gpointer) &found_desc,
							&found_null) )
		|| ( ( cmp == 3) && g_tree_lookup_extended(sstruct->strees[1],
							   d_struct, 
							   (gpointer) &found_desc,
							   &found_null) 
		     && g_tree_lookup_extended(sstruct->strees[2],
					       d_struct, 
					       (gpointer) &found_desc, 
					       &found_null) )
		|| ( ( cmp == 4) && g_tree_lookup_extended(sstruct->strees[1],
							   d_struct, 
							   (gpointer) &found_desc,
							   &found_null) 
		     && g_tree_lookup_extended(sstruct->strees[2], 
					       d_struct, 
					       (gpointer) &found_desc, 
					       &found_null) 
		     && g_tree_lookup_extended(sstruct->strees[3], 
					       d_struct, 
					       (gpointer) &found_desc,
					       &found_null) ) )
      {
	 // result found ... :-)
	 
#ifdef debug
	 if (found_desc != d_struct)
	    {
	       printf("\n\n[CRITICAL] descriptions should be identical!!");
	       printf("           %10d %s\n%10d %s\n\n",
		      (unsigned int) d_struct, d_struct->string, 
		      (unsigned int) found_desc, found_desc->string);
	    }
#endif
	 
	 sstruct->already_found++;
	 sstruct->current_desc = d_struct;
	 g_tree_foreach(d_struct->files, 
			(GTraverseFunc) iteration_add_desc_files, sstruct);
	 sstruct->current_desc = NULL;
      }
   

   // TODO: preferences settings

   //if (sstruct->already_found == max_tcp_results)
   //   return TRUE;
   //else
      return FALSE;
}





/* ************************************************************************** * 
 * now the stuff for the outside - these are the only ones which may be       *
 * called from files outside this one ...                                     *
 * ************************************************************************** */






/* ************************************************************************** * 
 * core functions                                                             *
 * ************************************************************************** */



int 
ed2k_data_init()
{
   
   rwlock = g_mutex_new();
   

   // user->id -> user
   UserTree = g_tree_new((GCompareFunc) user_id_order_func);

   // hash -> file
   FileTree = g_tree_new((GCompareFunc) filehash_order_func);

   // string -> desc
   // dont free the hash, its static
   //DescTree = g_tree_new((GCompareFunc) strcmp);

   // the description tree
   DescTree = g_tree_new((GCompareFunc) strcmp);

   // the words tree
   WordTree = g_tree_new((GCompareFunc) strcmp);

   // the low id list
   low_ids = g_new0(gpointer, MAX_LOW_ID_COUNT);
   
   num_users             = 0;
   num_files_all         = 0;
   num_files_different   = 0;
   size_all_files        = 0;

   printf("ed2k_data: initialized (rome = %s)\n", rome->string);


   return 1;
}




void
ed2k_data_read_lock()
{
   g_mutex_lock(rwlock);
}

void
ed2k_data_read_unlock()
{
   g_mutex_unlock(rwlock);
}

void
ed2k_data_write_lock()
{
   g_mutex_lock(rwlock);
}

void
ed2k_data_write_unlock()
{
   g_mutex_unlock(rwlock);
}




struct FileInfo *
ed2k_data_search_file_hash(char *hash)
{
   struct FileInfo          *file;
   gchar                    *orig_hash;
   gboolean                  found;
   
   
   found = g_tree_lookup_extended(FileTree, 
				  hash,
				  (gpointer*) &orig_hash,
				  (gpointer*) &file);
      
   if (found)
      return file;
   else 
      return NULL;
}



struct UserInfo *
ed2k_data_get_user_from_id(guint id)
{
   
   gboolean found;
   guint *found_id;
   struct UserInfo *user;
   
   found = g_tree_lookup_extended(UserTree, &(id), 
				  (gpointer) &found_id, 
				  (gpointer) &user);

   if (!found) 
      return NULL;
   
   g_assert(id == *found_id);
   g_assert(id ==  user->id);

   return user;
}




struct DescInfo *
ed2k_data_file_add_owner(struct FileInfo *file, struct UserInfo *owner,
			 char *name, unsigned short name_length)
{
   
   struct UserInfo *old_owner;
   struct DescInfo *old_desc;
   gboolean         owner_already_present;


   owner_already_present = g_tree_lookup_extended(file->owners, 
						  owner, 
						  (gpointer*) &old_owner,
						  (gpointer*) &old_desc);
						  
   if (owner_already_present)
      {
	 return NULL;
      }

   old_desc = description_increase_ref_count(name, name_length, file);
   g_tree_insert(file->owners, owner, old_desc); 
   g_tree_insert(owner->shared_files, file, old_desc);

   num_files_all++;
   file->count++;


   return old_desc;
}






struct FileInfo *
ed2k_data_file_new(char *hash, int size,
		   struct UserInfo *owner, 
		   char *name, unsigned short name_length)
{
   struct FileInfo *new_file;
   struct DescInfo *desc;


   // create struct
   new_file = g_new0(struct FileInfo, 1);

   // copy hash
   memcpy(new_file->hash, hash, 16);
   new_file->size   = size;
   new_file->owners = g_tree_new((GCompareFunc) simple_order_func);
   new_file->count  = 1;


   num_files_different++;
   num_files_all++;
   size_all_files += size;

   desc = description_increase_ref_count(name, name_length, new_file);

   g_tree_insert(FileTree, &(new_file->hash), new_file);
   g_tree_insert(new_file->owners, owner, desc); 
   g_tree_insert(owner->shared_files, new_file, desc);
   

   // return the thing
   return new_file;
}
   




struct FileInfo *
ed2k_data_file_add(char *hash, int size,
		   struct UserInfo *owner, 
		   char *name, unsigned short name_length)
{
   struct FileInfo *new_file;
   struct DescInfo *desc;
   
   struct DescInfo *found_desc;
   struct FileInfo *found_file;

   gboolean         found;


   new_file = ed2k_data_search_file_hash(hash);
   if (new_file)
      {
	 found = g_tree_lookup_extended(owner->shared_files, new_file,
					(gpointer*) &found_file, 
					(gpointer*) &found_desc);

	 if (found)
	    {
	       if (ndl(4))
		  printf("[PUB FAIL]      %16s has already published '%*s'\n", 
			 owner->inet_addr, name_length, name);
	       return 0;
	    }
	 else if (ndl(4))
	    printf("[PUBLISH<]      %16s offers: %*s\n", owner->inet_addr, 
		   name_length, name);
	 
	 new_file->count++;
      }


   if (! new_file)
      {
	 if (ndl(3))
	    printf("[PUBLISH<] %s offers: %*s\n", owner->inet_addr, 
		   name_length, name);

	 // create struct
	 new_file = g_new0(struct FileInfo, 1);
	 
	 // copy hash
	 memcpy(new_file->hash, hash, 16);
	 new_file->size   = size;
	 new_file->owners = g_tree_new((GCompareFunc) simple_order_func);
	 new_file->count  = 1;

	 num_files_different++;
	 size_all_files += size;
      }

   num_files_all++;
   
   desc = description_increase_ref_count(name, name_length, new_file);

   g_tree_insert(new_file->owners, owner, desc); 
   g_tree_insert(owner->shared_files, new_file, desc);
   g_tree_insert(FileTree, &(new_file->hash), new_file);


   // return the thing
   return new_file;
}
   


int 
ed2k_data_user_delete(struct UserInfo *user)
{
   struct UserInfo *user_lookup; 

   
   if (user->state != 1)
      return 0;


   user_lookup = search_user_id_internal(user->id);
   
   // nothing to do :-)
   if (!user_lookup)
      {
	 printf("[ ERROR! ] user not found! IP: %s, msg count: %d, size: %d\n",
		user->inet_addr, user->msg_count, user->msg_size);
	 ((rome->debugdata).user_not_found)++;
	 user->state = -1;
	 return 0;
      }
   
   // remove from tree ... 
   g_tree_remove(UserTree, &(user->id));    
   user->state = 0;
   num_users--;


#ifdef serious
   if (cvl() > 0)
      printf("[  drop  ] - %s:%d ver %d id %10u name '%s'\n", 
	     user->inet_addr, user->port, user->version, 
	     user->id, user->name);
#endif


   g_tree_foreach(user->shared_files, 
		  (GTraverseFunc)iteration_file_owner_remove, user);

   
   // so all left is to delete the tree 
   g_tree_destroy(user->shared_files);

   // decrease the number of users online
   
   // free a low id?
   if (user->id < 100000)
      {
	 free_low_id(user->id);
	 num_users_lowid--;
      }

   
   // TODO: logging




   return 1;
}




int
ed2k_data_user_new(struct UserInfo *user, 
		   gchar   *user_hash,
		   gchar   *user_name,
		   int      name_length,
		   gint     open)
{
   if (name_length > 29)
      name_length = 29;

   memcpy(user->hash, user_hash, 16);
   memcpy(user->name, user_name, name_length);
   user->shared_files = g_tree_new((GCompareFunc)simple_order_func);
   user->state        = 1;


   if (open)
      {
	 if (search_user_id_internal(user->id) )
	    {
	       ed2k_send_string(user, "your IP is already in use", -1);
	       ed2k_send_string(user, "giving you a low id", -1);
	       ed2k_send_string(user, "(are you behind a NAT firewall?)", -1);
	       user->id = reserve_low_id(get_next_low_id(), user);
	       num_users_lowid++;
	    }
	 else
	    user->id = user->ipaddr;
      }
   else
      {
	 user->id = reserve_low_id(get_next_low_id(), user);
	 num_users_lowid++;
      }

   g_assert(user->id > 0);

   if (user->state == -1) 
      {
	 printf("[ ERROR! ] user is already deleted!!!!\n");
	 rome->debugdata.doubledelete++;
	 return 0;
      }

   g_tree_insert(UserTree, &(user->id), user);
   user->state = 1;
   num_users++;


  return 1;
}





struct SearchResults *
ed2k_data_search_file_desc(char **words)
{
   
   struct SearchBaseStruct *sstruct;
   struct SearchResults    *sresults;
   GTree *gttemp;
   gint   snodes[5];
   gint   i, j;
   gpointer *word;
   gboolean  found;

   // quick'n dirty.

   // get a list of matching descriptions.
   // use only four words for searching


   sstruct = g_new0(struct SearchBaseStruct, 1);
   
   i = 0;
   found = TRUE;
   while (words[i] && found) 
   {
	 found = g_tree_lookup_extended(WordTree, words[i], 
					(gpointer) &word, 
					(gpointer) &(sstruct->strees[i]));
	 i++;
   }
   
   // one word not found :-) return nothing.
   if (!found) 
   {
	 return NULL;
   }

   for (j=0; j<i; j++)
      snodes[j] = g_tree_nnodes(sstruct->strees[j]);

   for (j=0; j<i-1; j++)
   {
	 if (snodes[j] > snodes[j+1])
	 {
	       gttemp               = sstruct->strees[j  ];
	       sstruct->strees[j  ] = sstruct->strees[j+1];
	       sstruct->strees[j+1] = gttemp;

	       snodes[4  ] = snodes[j  ];
	       snodes[j  ] = snodes[j+1];
	       snodes[j+1] = snodes[4  ];
	 }
   }

   // word search ended

   g_tree_foreach(sstruct->strees[0], 
		  (GTraverseFunc) iteration_search_desc, 
		  sstruct);
   

   // now sstruct->results is NULL or something. something is
   // good, NULL too. just return :-)
   

   sresults = sstruct->results;
   g_free(sstruct);
   
   return sresults;
}






struct SearchResults *
ed2k_data_generate_dl_sources(struct FileInfo *file)
{
   struct SearchBaseStruct  *results;
   struct SearchResults     *returnme;

   results = g_new0(struct SearchBaseStruct, 1);
   g_tree_foreach(file->owners, 
		  (GTraverseFunc) iteration_generate_dl_sources, results);

   returnme = results->results;
   g_free(results);
   
   return returnme;
}

