/* ************************************************************************** *
   This file is part of ceDonkey.

   ceDonkey is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   ceDonkey is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ceDonkey; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ************************************************************************** */
#define _MULTI_THREADED

#include <gnet/gnet.h>
#include <string.h>
#include <stdio.h>

#include "structs.h"
#include "ed2k_data.h"
#include "littlehelpers.h"
#include "ed2k_message_gen.h"
#include "serverlist.h"
#include "tcp_listeners.h"
#include "logging.h"
#include "settings.h"


/* locks */
GStaticRWLock  *rwlock;

/* stats */
extern struct central_struct *rome;




/* ************************************************************************** *
   some protocol backend functions                                      
   ************************************************************************** */

int 
handle_tag(struct TagRefStruct *tag, char *buffer)
{
   guint tag_length;

   switch ( *((char*)(buffer)) )
      {
      case 0x1:
	 printf("[WARNING] unimplemented master tag (0x01) !!\n");
	 g_assert_not_reached();
	 break;
      case 0x02:
	 tag->type   = 0x02;
	 tag->T1_len = *((gushort*)(buffer+1));
	 tag->T1     =  ((char*)(buffer+3));
	 tag->T2_len = *((gushort*)(buffer+3+tag->T1_len));
	 tag->T2     =  ((char*)(buffer+5+tag->T1_len));
	 tag_length  =  5 + tag->T1_len + tag->T2_len;
	 break;
      case 0x03:
	 tag->type   = 0x03;
	 tag->T1_len = *((gushort*)(buffer+1));
	 tag->T1     =  ((char*)(buffer+3));
	 tag->T2_len = 4;
	 tag->T2     =  ((char*)(buffer+5+tag->T1_len));
	 tag_length  = 7 + tag->T1_len;
	 break;
      default:
	 printf("[FATAL] unimplemented master tag %.1d !!\n", *((char*)(buffer)));
	 print_hex_dump(buffer, 32);
	 printf(".\n");
      }
   
   return tag_length;
}






/* ************************************************************************** *
   only pure packet processing functions here ... :-)                   
   ************************************************************************** */

gint ed2k_unknown_packet(struct message_handle_struct *message)
{
   if (message->udp_address)
      {
	 count_udp_traffic(U00, message->size, UDP_IN);
	 
	 fprintf(rome->log_files.unknown_udp,"UNK UDP: type %.2x size %d ", 
		 *((guchar*)((message->buffer)+1)), message->size);
	 fprintf(rome->log_files.unknown_udp,"from: %s:%d\n", 
		 gnet_inetaddr_get_canonical_name(message->udp_address),
		 gnet_inetaddr_get_port(message->udp_address));
	 write_hex_dump(rome->log_files.unknown_udp, message->buffer, message->size);
	 fprintf(rome->log_files.unknown_udp,"\n\n");
      }
   else
      {
	 count_tcp_traffic(NULL, T00, message->size);

	 fprintf(rome->log_files.unknown_tcp,"UNK TCP: type %.2x size %d ver %d ip %s:%d\n", 
		 *(unsigned short*)(message->buffer), message->size, 
		 message->client_info->version, 
		 message->client_info->inet_addr, 
		 message->client_info->port);
	 write_hex_dump(rome->log_files.unknown_tcp, message->buffer, message->size);
	 fprintf(rome->log_files.unknown_tcp,"\n\n");
      }

   return 0;
}




/* ************************************************************************** *
                                   UDP                                  
   ************************************************************************** */


// 0x96
gint ed2k_message_udp_ping(struct message_handle_struct *message)
{
   count_udp_traffic(U96, message->size, UDP_IN);
   log_write_0x96(message->udp_address, *((guint*)(message->buffer+2)), UDP_IN);

   ed2k_send_udp_pong(message);


   return 1;
}



// 0x97
gint ed2k_message_udp_pong(struct message_handle_struct *message)
{
   // we do only ping servers, so this has to be one ;-)

   count_udp_traffic(U97, message->size, UDP_IN);
   log_write_0x97(message->udp_address, *((guint*)(message->buffer+2)), UDP_IN);


   server_deregister_ping((GInetAddr *)message->udp_address, 
			  (guint*) ((message->buffer)+2));

   return 0;
}



// 0x9a
gint
ed2k_message_udp_query_sources(struct message_handle_struct *message)
{
   struct FileInfo       *file;
   struct SearchResults  *results;
   

   count_udp_traffic(U9a, message->size, UDP_IN);
   log_write_0x9a(message->udp_address, message->buffer+2);

   ed2k_data_read_lock();

   file = ed2k_data_search_file_hash(message->buffer + 2);

   if (!file)
      {
	 ed2k_data_read_unlock();
	 return 0;
      }

   results = ed2k_data_generate_dl_sources(file);

   ed2k_send_udp_download_sources((GInetAddr*) message->udp_address, 
				  (char*) &(file->hash), 
				  results, 0);

   ed2k_data_free_search_results(results);
   


   return 1;
}





// 0x9c push request 
gint
ed2k_message_udp_push_request(struct message_handle_struct *message)
{
   struct UserInfo *user_to_push;

   // what happens if both are firewalled? 

   count_udp_traffic(U9c, message->size, UDP_IN);
   log_write_0x9c(message->udp_address, message->buffer+2);


   user_to_push = ed2k_data_get_user_from_id( *((int*)((message->buffer)+8)) );
   
   if (user_to_push)
      ed2k_send_push_command(user_to_push, 
				 *((guint*)   (message->buffer)+2),
				 *((gushort*) (message->buffer)+6));
   else
      ed2k_send_udp_push_fail((GInetAddr*) message->udp_address);


   return 1;
}






// 0xa0
gint
ed2k_message_udp_srv_announce(struct message_handle_struct *message)
{
   count_udp_traffic(Ua0, message->size, UDP_IN);
   log_write_0xa0(message->udp_address, message->buffer+2, UDP_IN);
   
   server_lock();

   server_add_nn( *(int*)(message->buffer + 2), 
		  *(int*)(message->buffer + 6) );
   
   server_unlock();
   
   ed2k_send_udp_serverlist((GInetAddr*) message->udp_address);
   
   
   return 1;
}



// 0xa1, serverlist
gint
ed2k_message_udp_serverlist(struct message_handle_struct *message)
{
   // format: e3 a1 IP/PORT <num of following> IP/PORT, IP/PORT, ...

   int num_servers;
   int i;
   char *server;


   num_servers = *((unsigned char*)(message->buffer+8));

   count_udp_traffic(Ua1, message->size, UDP_IN);
   log_write_0xa1(message->udp_address, num_servers, UDP_IN);

   
   if (message->size != ((num_servers * 6) + 9) )
      {
	 fprintf(rome->log_files.buggy_pckts, 
		 "[BUGGY] received buggy 0xa1, UDP - wrong message size!\n");
	 fprintf(rome->log_files.buggy_pckts, 
		 "        size = %d, servers = %d\n", message->size, num_servers);
	 write_hex_dump(rome->log_files.buggy_pckts,  message->buffer, message->size);
	 fprintf(rome->log_files.buggy_pckts, "\n\n");
	 return 0;
      }

   server = message->buffer +2;
   server_add_nn(*((guint*)(server)), *((gushort*)(server+4)));

   server+=7;

   for (i=0; i<num_servers; i++)
      {
	 server_add_nn(*((guint*)(server)), *((gushort*)(server+4)));
	 server += 6;
      }
   
   return 1;
}







// TODO!! implement
// 0xa2, server information
gint
ed2k_message_udp_server_info(struct message_handle_struct *message)
{

   count_udp_traffic(Ua2, message->size, UDP_IN);
   // log_write_0xa2(message->udp_address, ..??.. , UDP_IN);

   
   return 1;
}




// 0xa4, serverlist request
gint
ed2k_message_udp_srvlist_request(struct message_handle_struct *message)
{
   count_udp_traffic(Ua4, message->size, UDP_IN);
   log_write_0xa4(message->udp_address, UDP_IN);

   ed2k_send_udp_serverlist(message->udp_address);
   
   return 1;
}




/* ************************************************************************** *
                                   TCP                                  
   ************************************************************************** */

// 0x01
gint ed2k_message_hello(gchar *buffer, gint size, struct UserInfo *ci)
{
   
   struct  TagRefStruct  tag;
   int                   user_is_open;
   guint                 tag_count,
                         i,
                         user_nlen,
                         user_cver,
                         user_port,
                         tag_value;
   guchar               *cur_tag,
                        *user_name,  
                        *user_hash;


   if (buffer[1] == 0x10) 
      buffer++;

   user_hash  = buffer+1;
   cur_tag    = buffer+27;
   tag_count  = *((int*)(buffer+23));
   
   count_tcp_traffic(ci, T01, size);

   for (i=0; i < tag_count; i++)
      {
	 tag_value =  *((guint*)(cur_tag));
	 
	 if (tag_value == TAG_CLIENT_VERSION)
	    {
	       user_cver = *((guint*)(cur_tag+4));
	       cur_tag += 8;
	    }
	 else if (tag_value == TAG_CLIENT_PORT)
	    {
	       user_port = *((guint*)(cur_tag+4));
	       cur_tag += 8;
	    }
	 else if (tag_value == TAG_USERNAME)
	    {
	       cur_tag += handle_tag(&tag, cur_tag);
	       user_nlen = tag.T2_len;
	       user_name = tag.T2;
	    }
	 else
	    {
	       printf("ed2k_message_hello: unexpected tag!\n");
	       cur_tag += handle_tag(&tag, cur_tag);
	    }
      }


   ci->listen_port = user_port;
   ci->version     = user_cver;

   // FUTURE: max. lowid connections have to be checked here

   // this function expects: 3 x  pointer, 1 x integer
   //                    ... 1 x boolean
   
   
   /*    ci->state = 10; */
   /*    user_is_open = tcp_client_is_open(ci); */
   /*    if (user_is_open) */
   /*       ci->state = 1; */
   /*    else */
   /*       ci->state = 2; */

   // all are open for now

   ci->state = 1;
   user_is_open = 1;

   ed2k_data_write_lock();
   ed2k_data_user_new(ci, user_hash, user_name, user_nlen, user_is_open);
   ed2k_data_write_unlock();

   log_write_0x01(ci);

   // condition: hello packet only sent once, is answered by a 0x40
   ed2k_send_0x40(ci);
   ed2k_send_server_status(ci);
   ed2k_send_server_greeting(ci);
   

   return 1;
}



// 0x15
// tested with protocol version 59
gint ed2k_message_publish_files(gchar *buffer, gint size, struct UserInfo *ci)
{
   struct TagRefStruct tag;
   unsigned short num_files;
   int i, j;
   char     *offset;
   guint     file_size,
             tag_value, 
             tag_count;
   gchar    *file_name;
   gushort   name_size;
   gchar    *file_hash;
   
   // TODO: mutexes ... 


   count_tcp_traffic(ci, T15, size);

   num_files = *((unsigned int*) (buffer+1));



   ci->shared_files_count += num_files;
   
   offset = buffer + 5;

	 // well, DONT USE OFFSET AFTER THIS POINT!!!
	 // IF OFFSET POINTS TO SOMETHING INTERESTING, JUST SAVE
	 // IT AND CONTINUE, BUT DONT REFERENCE TO OFFSET ITSELF!!!

   ed2k_data_write_lock();

   for (i = 0; i < num_files; i++)
      {

	 file_hash = offset;
	 offset += 22;
	 tag_count = *((guint*)(offset));
	 offset +=  4;

	 for (j = 0; j < tag_count; j++)
	    {
	       tag_value = *((guint*)(offset));
	       if (tag_value == TAG_FILE_NAME)
		  {
		     offset    += handle_tag(&tag, offset);
		     file_name  = tag.T2;
		     name_size  = tag.T2_len;
		  }
	       else if (tag_value == TAG_FILE_SIZE)
		  {
		     file_size  = *((guint*)(offset + 4));
		     offset    += 8;
		  }
	       else if (tag_value == TAG_FILE_AVAIL)
		  {
		     // senseless in publish files!
		     offset    += 8;
		  }
	       else
		  {
		     offset += handle_tag(&tag, offset);
		  }
	    }

	 // if new create & add owner, else just add owner
	 
	 ed2k_data_file_add(file_hash, file_size, ci, file_name, name_size);

      }
   

   ed2k_data_write_unlock();

   
   return 1;
}


// 0x16
gint ed2k_message_search_file_desc(gchar *buffer, gint size, struct UserInfo *ci)
{
   unsigned short len_string, len_type;
   unsigned int type;
   char *typestring, *searchstring;
   struct SearchResults *files;
   time_t cur_time; 
   char **argv;


   count_tcp_traffic(ci, T16, size);

   if ( *((char*)(buffer+1)) == 1 )
      {
	 // unspecific search
	 type = 0;
	 searchstring = (buffer + 4);
	 len_string = *((unsigned short*)(buffer+2));
      }
   else 
      {
	 len_type = *((unsigned short*)(buffer+4));
	 typestring = buffer + 6;
	 len_string = *((unsigned short*)(buffer+10+len_type));
	 printf("%d", len_string);
	 searchstring = buffer + 12 + len_type;

	 type = 1;
      }

   
   if (ndl(4)) 
      printf("[SEARCH] U:%s S:'%*s'\n", ci->inet_addr, len_string, searchstring);


   if (rome->settings.log_everything || rome->settings.log_text_searches_tcp)
      {
	 cur_time = time(NULL);
	 fprintf(rome->log_files.searches, 
		 "%12d 0x16 U:%s S:'%*s'\n", 
		 (unsigned int) cur_time,
		 ci->inet_addr, len_string, searchstring);

	 if (rome->settings.log_everything) 
	    fprintf(rome->log_files.trace, 
		    "%12d 0x16 U:%s S:'%*s'\n", 
		    (unsigned int) cur_time,
		    ci->inet_addr, len_string, searchstring);
      }
	 

   argv = parse_string(searchstring, len_string);

   ed2k_data_read_lock();

   files = ed2k_data_search_file_desc(argv);
   free_parsed_strings(argv);
   
   // also when no results available!! (and delete lock in there)
   ed2k_return_search_results(ci, files, 0);

   ed2k_data_free_search_results(files);


   return 1;
}





// 0x19
gint ed2k_message_query_sources(gchar *buffer, gint size, struct UserInfo *ci)
{
   struct FileInfo       *file;
   struct SearchResults  *results;

   count_tcp_traffic(ci, T19, size);
   

   ed2k_data_read_lock();

   file = ed2k_data_search_file_hash(buffer + 1);

   if (!file) 
      {
	 ed2k_data_read_unlock();
	 return 0;
      }
   
   results = ed2k_data_generate_dl_sources(file);

   ed2k_send_download_sources(ci, file, results, 0);

   ed2k_data_free_search_results(results);


   return 1;
}




// 0x1c
gint
ed2k_message_push_request(gchar *buffer, gint size, struct UserInfo *ci)
{
   struct UserInfo *user_to_push;
   char             ipaddr[16];
   time_t           cur_time;
   

   // what happens if both are firewalled? 

   count_tcp_traffic(ci, T1c, size);

   if (ndl(4))
      {
	 inet_ntoc(*((int*)(buffer+1)), ipaddr);
	 printf("[PUSHREQ] F:%s T:%s\n",
		ci->inet_addr,
		ipaddr);
      }	 

   if (rome->settings.log_everything) 
      {
	 cur_time = time(NULL);
	 inet_ntoc(*((int*)(buffer+1)), ipaddr);
	 fprintf(rome->log_files.trace, "%12d 0x1c F:%s T:%s\n",
		 (unsigned  int) cur_time,
		 ci->inet_addr,
		 ipaddr);	 
      }

   ed2k_data_read_lock();

   user_to_push = ed2k_data_get_user_from_id( *((int*)(buffer+1)) );
   
   if (user_to_push && user_to_push->id > 1000000)
      ed2k_send_push_command(user_to_push, ci->id, ci->listen_port);
   else
      ed2k_send_push_fail(ci);

   ed2k_data_read_unlock();

   return 1;
}






gint
ed2k_message_handler_init(struct central_struct *_rome)
{
   
   return 1;
}





/* ************************************************************************** *
 * ed2k_handle_message()                                                      *
 * -------------------------------------------------------------------------- *
 * this function is the base function of the thread pool, which takes care    *
 * of the packet processing. this should be all there is needed :-)           *
 * ************************************************************************** */

// gint ed2k_handle_packet(char* buffer, gint size, struct UserInfo *client_info)
gint 
ed2k_handle_message(struct message_handle_struct *message)
{
   //printf("[ED2k ] processing packet type: 0x%.2x\n", type);


   gchar            *buffer;
   // see [4]
   guchar            type;
   guint             size;
   struct UserInfo  *client_info;



   size = message->size;
   buffer = message->buffer;
   client_info = message->client_info;


   // see ARCHITECTURE [3] & [6]
   if (message->udp_address)
      {
	 if (! rome->settings.public)
	    return 0;
	 type = buffer[1];
      }
   else 
      {
	 type = buffer[0];
	 if (message->client_info->state != 1 
	     && message->client_info->state != 2
	     && type != 0x01)
	    return 0;
      }


   switch (type)
      {
	 
	 // tcp tests first :-)
	 
      case 0x01:
	 // hello packet
	 ed2k_message_hello(buffer, size, client_info);
	 ed2k_send_server_desc(client_info);
	 break;
      case 0x14:
	 // get server list
	 ed2k_send_serverlist(client_info);
	 break;
      case 0x15:
	 // files offered packet
	 ed2k_message_publish_files(buffer, size, client_info);
	 break;
      case 0x16:
	 // search file packet
	 ed2k_message_search_file_desc(buffer, size, client_info);
	 break;
      case 0x19:
	 // query download sources
	 ed2k_message_query_sources(buffer, size, client_info);
	 break;
      case 0x1c:
	 // push request
	 ed2k_message_push_request(buffer, size, client_info);
	 break;


	 // udp things later :-)


      case 0x96:
	 // PING 
	 ed2k_message_udp_ping(message);
	 break;
      case 0x97:
	 // PONG
	 ed2k_message_udp_pong(message);
	 break;
      case 0x9a:
	 // query sources
	 ed2k_message_udp_query_sources(message);
	 break;
      case 0xa0:
	 // announce server
	 ed2k_message_udp_srv_announce(message);
	 break;
      case 0xa1:
	 // serverlist
	 ed2k_message_udp_serverlist(message);
	 break;
      case 0xa2:
	 // server information ??
	 ed2k_message_udp_server_info(message);
	 break;
      case 0xa4:
	 // request serverlist
	 ed2k_message_udp_srvlist_request(message);
	 break;

	 // default ...


      default:
	 ed2k_unknown_packet(message);
	 break;

      }


   return 1;
}




gint ed2k_message_popper(struct message_handle_struct *message, 
				gpointer data)
{
   ed2k_handle_message(message);
   ed2k_data_free_message(message);
   return 1;
}
